"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

from ast import arguments
import logging
from urllib.parse import urlparse
from opentelemetry.sdk.trace import Span
from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    get_keys_as_tuple,
    get_nested_value,
    try_option,
    get_exception_message,
    get_status_code,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_llamaindex_finish_reason_to_finish_type

LLAMAINDEX_AGENT_NAME_KEY = "_active_agent_name"
logger = logging.getLogger(__name__)

def get_status(result):
    if result is not None and hasattr(result, 'status'):
        return result.status
    return None

def extract_tools(instance):
    tools = []
    if not hasattr(instance, 'state') or not hasattr(instance.state, 'task_dict'):
        return []
    try:
        data = next(iter(instance.state.task_dict.values())).task
    except (AttributeError, StopIteration):
        return []

    if hasattr(data,'extra_state') and 'sources' in data.extra_state:
        for tool_output in data.extra_state['sources']:
            tool_name = tool_output.tool_name
            if tool_name:
                tools.append(tool_name)
    return tools

def get_tool_name(args, instance):
    if len(args) > 1:
        if hasattr(args[1], 'metadata') and hasattr(args[1].metadata, 'name'):
            return args[1].metadata.name
        return ""
    else:
        if hasattr(instance, 'metadata') and hasattr(instance.metadata, 'name'):
            return instance.metadata.name
        return ""

def get_tool_description(arguments):
    if len(arguments['args']) > 1:
        if hasattr(arguments['args'][1], 'metadata') and hasattr(arguments['args'][1].metadata, 'description'):
            return arguments['args'][1].metadata.description
        return ""
    else:
        if hasattr(arguments['instance'], 'metadata') and hasattr(arguments['instance'].metadata, 'description'):
            return arguments['instance'].metadata.description
        return ""

def extract_tool_args(arguments):
    tool_args = []
    if len(arguments['args']) > 1:
        for key, value in arguments['args'][2].items():
            # check if value is builtin type or a string
            if value is not None and isinstance(value, (str, int, float, bool)):
                tool_args.append({key, value})
    else:
        for key, value in arguments['kwargs'].items():
            # check if value is builtin type or a string
            if value is not None and isinstance(value, (str, int, float, bool)):
                tool_args.append({key, value})
    return [get_json_dumps(tool_arg) for tool_arg in tool_args]

def extract_tool_response(response):
    if hasattr(response, 'raw_output'):
        return response.raw_output
    return ""

def is_delegation_tool(args, instance) -> bool:
    return get_tool_name(args, instance) == "handoff"

def get_agent_name(instance) -> str:
    if hasattr(instance, 'name'):
        return instance.name
    else:
        return instance.__class__.__name__

def get_agent_description(instance) -> str:
    if hasattr(instance, 'description'):
        return instance.description
    return ""

def get_source_agent(parent_span:Span) -> str:
    source_agent_name = parent_span.attributes.get(LLAMAINDEX_AGENT_NAME_KEY, "")
    if source_agent_name == "" and parent_span.name.startswith("llama_index.core.agent.ReActAgent."):
        # Fallback to the agent name from the parent span if not set
        source_agent_name = "ReactAgent"
    return source_agent_name

def get_target_agent(results) -> str:
    if hasattr(results, 'raw_input'):
        return results.raw_input.get('kwargs', {}).get("to_agent", "")
    return ""

def extract_messages(args):
    """Extract system and user messages"""
    try:
        messages = []

        def process_message(msg):
            """Processes a single message and extracts relevant information."""
            if hasattr(msg, 'content') and hasattr(msg, 'role'):
                role = getattr(msg.role, 'value', msg.role)
                content = msg.content if role == "system" else extract_query_from_content(msg.content)
                messages.append({role: content})

        if isinstance(args, (list, tuple)) and args:
            for msg in args[0]:
                process_message(msg)
        elif args and isinstance(args, tuple):
            messages.append(args[0])
        if isinstance(args, dict):
            for msg in args.get("messages", []):
                process_message(msg)
        

        return [get_json_dumps(message) for message in messages]

    except Exception as e:
        logger.warning("Error in extract_messages: %s", str(e))
        return []

def extract_agent_input(args):
    if isinstance(args, (list, tuple)):
        input_args = []
        for arg in args:
            if isinstance(arg, (str, dict)):
                input_args.append(arg)
            elif hasattr(arg, 'raw') and isinstance(arg.raw, str):
                input_args.append(arg.raw)
        return input_args
    elif isinstance(args, str):
        return [args]
    return ""

def extract_agent_response(arguments):
    status = get_status_code(arguments)
    if status == 'success':
        if hasattr(arguments['result'], 'response'):
            if hasattr(arguments['result'].response, 'content'):
                return arguments['result'].response.content
            return arguments['result'].response
        return ""
    else:
        if arguments["exception"] is not None:
            return get_exception_message(arguments)
        elif hasattr(arguments['result'], "error"):
            return arguments['result'].error

def extract_assistant_message(arguments):
    status = get_status_code(arguments)
    messages = []
    role = "assistant"
    if status == 'success':
        if isinstance(arguments['result'], str):
            messages.append({role: arguments['result']})
        if hasattr(arguments['result'], "content"):
            messages.append({role: arguments['result'].content})
        if hasattr(arguments['result'], "message") and hasattr(arguments['result'].message, "content"):
            role = getattr(arguments['result'].message, 'role', role)
            if hasattr(role, 'value'):
                role = role.value
            messages.append({role: arguments['result'].message.content})
        if hasattr(arguments['result'],"response") and isinstance(arguments['result'].response, str):
            messages.append({role: arguments['result'].response})
    else:
        if arguments["exception"] is not None:
            return get_exception_message(arguments)
        elif hasattr(arguments['result'], "error"):
            return arguments['result'].error

    return get_json_dumps(messages[0]) if messages else ""

def extract_query_from_content(content):
    try:
        query_prefix = "Query:"
        answer_prefix = "Answer:"
        query_start = content.find(query_prefix)
        if query_start == -1:
            return content

        query_start += len(query_prefix)
        answer_start = content.find(answer_prefix, query_start)
        if answer_start == -1:
            query = content[query_start:].strip()
        else:
            query = content[query_start:answer_start].strip()
        return query
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_query_from_content: %s", str(e))
        return ""


def extract_provider_name(instance):
    if hasattr(instance,'api_base'):
        provider_url: Option[str]= try_option(getattr, instance, 'api_base').and_then(lambda url: urlparse(url).hostname)
    if hasattr(instance,'_client'):
        provider_url:Option[str] = try_option(getattr, instance._client.base_url,'host')
    if hasattr(instance, 'model') and isinstance(instance.model, str) and 'gemini' in instance.model.lower():
        provider_url: Option[str] = try_option(lambda: 'gemini.googleapis.com')
    return provider_url.unwrap_or(None)


def extract_inference_endpoint(instance):
    if hasattr(instance,'_client'):
        if hasattr(instance._client,'sdk_configuration'):
            inference_endpoint: Option[str] = try_option(getattr, instance._client.sdk_configuration, 'server_url').map(str)
        if hasattr(instance._client,'base_url'):
            inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if hasattr(instance, 'model') and isinstance(instance.model, str) and 'gemini' in instance.model.lower():
        inference_endpoint = try_option(lambda: f"https://generativelanguage.googleapis.com/v1beta/models/{instance.model}:generateContent")
    return inference_endpoint.unwrap_or(extract_provider_name(instance))


def extract_vectorstore_deployment(my_map):
    if isinstance(my_map, dict):
        if '_client_settings' in my_map:
            client = my_map['_client_settings'].__dict__
            host, port = get_keys_as_tuple(client, 'host', 'port')
            if host:
                return f"{host}:{port}" if port else host
        keys_to_check = ['client', '_client']
        host = __get_host_from_map(my_map, keys_to_check)
        if host:
            return host
    else:
        if hasattr(my_map, 'client') and '_endpoint' in my_map.client.__dict__:
            return my_map.client.__dict__['_endpoint']
        host, port = get_keys_as_tuple(my_map.__dict__, 'host', 'port')
        if host:
            return f"{host}:{port}" if port else host
    return None


def __get_host_from_map(my_map, keys_to_check):
    for key in keys_to_check:
        seed_connections = get_nested_value(my_map, [key, 'transport', 'seed_connections'])
        if seed_connections and 'host' in seed_connections[0].__dict__:
            return seed_connections[0].__dict__['host']
    return None


def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None


def update_input_span_events(args):
    if isinstance(args, tuple):
        return args[0].query_str if len(args) > 0 else ""


def update_output_span_events(results):
    if isinstance(results, list) and len(results) >0:
        output_arg_text = results[0].text
        if len(output_arg_text) > 100:
            output_arg_text = output_arg_text[:100] + "..."
        return output_arg_text


def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and hasattr(response, "raw"):
        if response.raw is not None:
            token_usage = response.raw.get("usage") if isinstance(response.raw, dict) else getattr(response.raw, "usage", None)
            if token_usage is None:
                token_usage = response.raw.get("usage_metadata") if isinstance(response.raw, dict) else getattr(response.raw,
                                                                                                       "usage_metadata", None)
            if token_usage is not None:
                temperature = instance.__dict__.get("temperature", None)
                meta_dict.update({"temperature": temperature})
                meta_dict.update({"completion_tokens": getattr(token_usage, "completion_tokens",None) or getattr(token_usage,"output_tokens",None) or token_usage.get("candidates_token_count",None)})
                meta_dict.update({"prompt_tokens": getattr(token_usage, "prompt_tokens",None) or getattr(token_usage,"input_tokens",None) or token_usage.get("prompt_token_count",None)})
                total_tokens = getattr(token_usage, "total_tokens", None)
                if total_tokens is not None:
                    meta_dict.update({"total_tokens": total_tokens})
                else:
                    output_tokens = getattr(token_usage, "output_tokens", None)
                    input_tokens = getattr(token_usage, "input_tokens", None)
                    if output_tokens is not None and input_tokens is not None:
                        meta_dict.update({"total_tokens": output_tokens + input_tokens})
                    else:
                        meta_dict.update({ "total_tokens": token_usage.get("total_token_count", None)})

    return meta_dict

def extract_finish_reason(arguments):
    """Extract finish_reason from LlamaIndex response."""
    try:
        # Handle exception cases first
        if arguments.get("exception") is not None:
            return "error"
        
        response = arguments.get("result")
        if response is None:
            return None
            
        # Check various possible locations for finish_reason in LlamaIndex responses
        
        # Direct finish_reason attribute
        if hasattr(response, "finish_reason") and response.finish_reason:
            return response.finish_reason
            
        # Check if response has raw attribute (common in LlamaIndex)
        if hasattr(response, "raw") and response.raw:
            raw_response = response.raw
            if isinstance(raw_response, dict):
                # Check for finish_reason in raw response
                if "finish_reason" in raw_response:
                    return raw_response["finish_reason"]
                if "stop_reason" in raw_response:
                    return raw_response["stop_reason"]
                # Check for choices structure (OpenAI-style)
                if "choices" in raw_response and raw_response["choices"]:
                    choice = raw_response["choices"][0]
                    if isinstance(choice, dict) and "finish_reason" in choice:
                        return choice["finish_reason"]
            elif hasattr(raw_response, "choices") and raw_response.choices:
                # Handle object-style raw response
                choice = raw_response.choices[0]
                if hasattr(choice, "finish_reason"):
                    return choice.finish_reason
        
        # Check for additional metadata
        if hasattr(response, "additional_kwargs") and response.additional_kwargs:
            kwargs = response.additional_kwargs
            if isinstance(kwargs, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in kwargs:
                        return kwargs[key]
        
        # Check for response metadata
        if hasattr(response, "response_metadata") and response.response_metadata:
            metadata = response.response_metadata
            if isinstance(metadata, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in metadata:
                        return metadata[key]
        
        # Check for source nodes or other LlamaIndex-specific attributes
        if hasattr(response, "source_nodes") and response.source_nodes:
            # If we have source nodes, it's likely a successful retrieval
            return "stop"
        
        # If no specific finish reason found, infer from status
        status_code = get_status_code(arguments)
        if status_code == 'success':
            return "stop"  # Default success finish reason
        elif status_code == 'error':
            return "error"
            
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    
    return None


def map_finish_reason_to_finish_type(finish_reason):
    """Map LlamaIndex finish_reason to finish_type."""
    return map_llamaindex_finish_reason_to_finish_type(finish_reason)

def extract_agent_request_input(kwargs):
    if "user_msg" in kwargs:
        return kwargs["user_msg"]
    return ""

def extract_agent_request_output(arguments):
    if hasattr(arguments['result'], 'response'):
        if hasattr(arguments['result'].response, 'content'):
            return arguments['result'].response.content
        return arguments['result'].response
    elif hasattr(arguments['result'], 'raw_output'):
        return arguments['result'].raw_output
    return ""



from monocle_apptrace.instrumentation.metamodel.mcp import _helper

TOOLS = {
    "type": "agentic.mcp.invocation",
    "attributes": [
        [
            {
                "_comment": "name of the tool",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_name(arguments),
            },
            {
                "_comment": "tool type",
                "attribute": "type",
                "accessor": lambda arguments: "mcp.server",
            },
            {
                "_comment": "tool url",
                "attribute": "url",
                "accessor": lambda arguments: _helper.get_url(arguments),
            },
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is Tool input",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.get_params_arguments(
                        arguments
                    ),
                },
            ],
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is Tool output",
                    "attribute": "output",
                    "accessor": lambda arguments: _helper.get_output_text(arguments)
                },
            ],
        },
    ],
}



from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import resolve_from_alias
import logging
import json

logger = logging.getLogger(__name__)

def get_url(arguments):
    """Get the URL of the tool from the instance."""
    return arguments["instance"].url

def get_method(arguments):
    """Get the method of the tool from the instance."""
    return arguments["args"][0].method

def get_params_arguments(arguments):
    """Get the params of the tool from the instance."""
    return arguments["args"][0].params.message.parts[0].root.text

def get_role(arguments):
    """Get the role of the tool from the instance."""
    return arguments["args"][0].params.message.role.value

def get_status(arguments):
    """Get the status of the tool from the result."""
    return arguments["result"].root.result.status.state.value

def get_response(arguments):
    """Get the response of the tool from the result."""
    ret_val = []
    for artifact in arguments["result"].root.result.artifacts:
        if artifact.parts:
            for part in artifact.parts:
                if part.root.text:
                    ret_val.append(part.root.text)
    return ret_val
    # return arguments["result"].root.result.artifacts[0].parts[0].root.text



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper
from monocle_apptrace.instrumentation.metamodel.botocore.entities.inference import (
    INFERENCE,
)

BOTOCORE_METHODS = [
    {
      "package": "botocore.client",
      "object": "ClientCreator",
      "method": "create_client",
      "wrapper_method": task_wrapper,
      "span_handler":"botocore_handler",
      "output_processor": INFERENCE,
      "skip_span": True
    }
]


"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import logging
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    get_keys_as_tuple,
    get_nested_value,
    try_option,
    get_exception_message,
    get_status_code,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_langchain_finish_reason_to_finish_type


logger = logging.getLogger(__name__)


def extract_messages(args):
    """Extract system and user messages"""
    try:
        messages = []
        if args and isinstance(args, (list, tuple)) and hasattr(args[0], 'text'):
            return [args[0].text]
        if args and isinstance(args, (list, tuple)) and len(args) > 0:
            if isinstance(args[0], list) and len(args[0]) > 0:
                first_msg = args[0][0]
                if hasattr(first_msg, 'content') and hasattr(first_msg, 'type') and first_msg.type == "human":
                    return args[0][0].content
            if hasattr(args[0], "messages") and isinstance(args[0].messages, list):
                for msg in args[0].messages:
                    if hasattr(msg, 'content') and hasattr(msg, 'type'):
                        messages.append({msg.type: msg.content})
            else:
                for msg in args[0]:
                    if hasattr(msg, 'content') and hasattr(msg, 'type') and msg.content:
                        messages.append({msg.type: msg.content})
                    elif hasattr(msg, 'tool_calls') and msg.tool_calls:
                        messages.append({msg.type: get_json_dumps(msg.tool_calls)})
        return [get_json_dumps(d) for d in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def extract_assistant_message(arguments):
    status = get_status_code(arguments)
    messages = []
    role = "assistant"
    if status == 'success':
        if isinstance(arguments['result'], str):
            messages.append({role: arguments['result']})
        elif hasattr(arguments['result'], "content") and arguments['result'].content != "":
            role = arguments['result'].type if hasattr(arguments['result'], 'type') else role
            messages.append({role: arguments['result'].content})
        elif hasattr(arguments['result'], "message") and hasattr(arguments['result'].message, "content") and arguments['result'].message.content != "":
            role = arguments['result'].type if hasattr(arguments['result'], 'type') else role
            messages.append({role: arguments['result'].message.content})
        elif hasattr(arguments['result'], "tool_calls"):
            role = arguments['result'].type if hasattr(arguments['result'], 'type') else role
            messages.append({role: arguments['result'].tool_calls[0]})
    else:
        if arguments["exception"] is not None:
            messages.append({role: get_exception_message(arguments)})
        elif hasattr(arguments["result"], "error"):
            return arguments["result"].error
    return get_json_dumps(messages[0]) if messages else ""

def extract_provider_name(instance):
    provider_url: Option[str] = Option(None)
    if hasattr(instance, 'client'):
        provider_url: Option[str] = try_option(getattr, instance.client, 'universe_domain')
    if hasattr(instance,'client') and hasattr(instance.client, '_client') and hasattr(instance.client._client, 'base_url'):
        # If the client has a base_url, extract the host from it
        provider_url: Option[str] = try_option(getattr, instance.client._client.base_url, 'host')
    if hasattr(instance, '_client') and hasattr(instance._client, 'base_url'):
        provider_url = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)


def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = None
    # instance.client.meta.endpoint_url
    if hasattr(instance, 'client') and hasattr(instance.client, 'transport'):
        inference_endpoint: Option[str] = try_option(getattr, instance.client.transport, 'host')

    if hasattr(instance, 'client') and hasattr(instance.client, 'meta') and hasattr(instance.client.meta, 'endpoint_url'):
        inference_endpoint: Option[str] = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    if hasattr(instance,'client') and hasattr(instance.client, '_client'):
        inference_endpoint: Option[str] = try_option(getattr, instance.client._client, 'base_url').map(str)

    if hasattr(instance,'_client'):
        inference_endpoint = try_option(getattr, instance._client, 'base_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))


def extract_vectorstore_deployment(my_map):
    if isinstance(my_map, dict):
        if '_client_settings' in my_map:
            client = my_map['_client_settings'].__dict__
            host, port = get_keys_as_tuple(client, 'host', 'port')
            if host:
                return f"{host}:{port}" if port else host
        keys_to_check = ['client', '_client']
        host = __get_host_from_map(my_map, keys_to_check)
        if host:
            return host
    else:
        if hasattr(my_map, 'client') and '_endpoint' in my_map.client.__dict__:
            return my_map.client.__dict__['_endpoint']
        host, port = get_keys_as_tuple(my_map.__dict__, 'host', 'port')
        if host:
            return f"{host}:{port}" if port else host
    return None

def __get_host_from_map(my_map, keys_to_check):
    for key in keys_to_check:
        seed_connections = get_nested_value(my_map, [key, 'transport', 'seed_connections'])
        if seed_connections and 'host' in seed_connections[0].__dict__:
            return seed_connections[0].__dict__['host']
    return None

def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None


def update_input_span_events(args):
    return args[0] if len(args) > 0 else ""


def update_output_span_events(results):
    output_arg_text = " ".join([doc.page_content for doc in results if hasattr(doc, 'page_content')])
    if len(output_arg_text) > 100:
        output_arg_text = output_arg_text[:100] + "..."
    return output_arg_text


def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and hasattr(response, "response_metadata"):
        if hasattr(response, "usage_metadata") and response.usage_metadata is not None:
            token_usage = response.usage_metadata
        else:
            response_metadata = response.response_metadata
            token_usage = response_metadata.get("token_usage")
        if token_usage is not None:
            temperature = instance.__dict__.get("temperature", None)
            meta_dict.update({"temperature": temperature})
            meta_dict.update(
                {"completion_tokens": token_usage.get("completion_tokens") or token_usage.get("output_tokens")})
            meta_dict.update({"prompt_tokens": token_usage.get("prompt_tokens") or token_usage.get("input_tokens")})
            meta_dict.update({"total_tokens": token_usage.get("total_tokens")})
    return meta_dict

def extract_finish_reason(arguments):
    """Extract finish_reason from LangChain response."""
    try:
        # Handle exception cases first
        if arguments.get("exception") is not None:
            # If there's an exception, it's typically an error finish type
            return "error"
        
        response = arguments.get("result")
        if response is None:
            return None
            
        # Check various possible locations for finish_reason in LangChain responses
        
        # Direct finish_reason attribute
        if hasattr(response, "finish_reason") and response.finish_reason:
            return response.finish_reason
            
        # Response metadata (common in LangChain)
        if hasattr(response, "response_metadata") and response.response_metadata:
            metadata = response.response_metadata
            if isinstance(metadata, dict):
                # Check for finish_reason in metadata
                if "finish_reason" in metadata:
                    return metadata["finish_reason"]
                # Check for stop_reason (Anthropic style through LangChain)
                if "stop_reason" in metadata:
                    return metadata["stop_reason"]
                # Check for other common finish reason keys
                for key in ["completion_reason", "end_reason", "status"]:
                    if key in metadata:
                        return metadata[key]
        
        # Check if response has generation_info (some LangChain models)
        if hasattr(response, "generation_info") and response.generation_info:
            gen_info = response.generation_info
            if isinstance(gen_info, dict):
                for key in ["finish_reason", "stop_reason", "completion_reason"]:
                    if key in gen_info:
                        return gen_info[key]
        
        # Check if response has llm_output (batch responses)
        if hasattr(response, "llm_output") and response.llm_output:
            llm_output = response.llm_output
            if isinstance(llm_output, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in llm_output:
                        return llm_output[key]
        
        # For AIMessage responses, check additional_kwargs
        if hasattr(response, "additional_kwargs") and response.additional_kwargs:
            kwargs = response.additional_kwargs
            if isinstance(kwargs, dict):
                for key in ["finish_reason", "stop_reason"]:
                    if key in kwargs:
                        return kwargs[key]
        
        # For generation responses with choices (similar to OpenAI structure)
        if hasattr(response, "generations") and response.generations:
            generations = response.generations
            if isinstance(generations, list) and len(generations) > 0:
                for generation in generations:
                    if hasattr(generation, "generation_info") and generation.generation_info:
                        gen_info = generation.generation_info
                        if isinstance(gen_info, dict):
                            for key in ["finish_reason", "stop_reason"]:
                                if key in gen_info:
                                    return gen_info[key]
        
        # If no specific finish reason found, infer from status
        status_code = get_status_code(arguments)
        if status_code == 'success':
            return "stop"  # Default success finish reason
        elif status_code == 'error':
            return "error"
            
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    
    return None


def map_finish_reason_to_finish_type(finish_reason):
    """Map LangChain finish_reason to finish_type."""
    return map_langchain_finish_reason_to_finish_type(finish_reason)


# Azure environment constants
AZURE_ML_ENDPOINT_ENV_NAME = "AZUREML_ENTRY_SCRIPT"
AZURE_FUNCTION_WORKER_ENV_NAME = "FUNCTIONS_WORKER_RUNTIME"
AZURE_APP_SERVICE_ENV_NAME = "WEBSITE_SITE_NAME"
AWS_LAMBDA_ENV_NAME = "AWS_LAMBDA_RUNTIME_API"
GITHUB_CODESPACE_ENV_NAME = "CODESPACES"

AWS_LAMBDA_FUNCTION_IDENTIFIER_ENV_NAME = "AWS_LAMBDA_FUNCTION_NAME"
AZURE_FUNCTION_IDENTIFIER_ENV_NAME = "WEBSITE_SITE_NAME"
AZURE_APP_SERVICE_IDENTIFIER_ENV_NAME = "WEBSITE_DEPLOYMENT_ID"
GITHUB_CODESPACE_IDENTIFIER_ENV_NAME = "GITHUB_REPOSITORY"


# Azure naming reference can be found here
# https://learn.microsoft.com/en-us/azure/cloud-adoption-framework/ready/azure-best-practices/resource-abbreviations
# https://docs.aws.amazon.com/resource-explorer/latest/userguide/supported-resource-types.html#services-lookoutmetrics
AZURE_FUNCTION_NAME = "azure_func"
AZURE_APP_SERVICE_NAME = "azure_webapp"
AZURE_ML_SERVICE_NAME = "azure_ml"
AWS_LAMBDA_SERVICE_NAME = "aws_lambda"
GITHUB_CODESPACE_SERVICE_NAME = "github_codespace"

# Env variables to identify infra service type
service_type_map = {
    AZURE_ML_ENDPOINT_ENV_NAME: AZURE_ML_SERVICE_NAME,
    AZURE_APP_SERVICE_ENV_NAME: AZURE_APP_SERVICE_NAME,
    AZURE_FUNCTION_WORKER_ENV_NAME: AZURE_FUNCTION_NAME,
    AWS_LAMBDA_ENV_NAME: AWS_LAMBDA_SERVICE_NAME,
    GITHUB_CODESPACE_ENV_NAME: GITHUB_CODESPACE_SERVICE_NAME
}

# Env variables to identify infra service name
service_name_map = {
    AZURE_APP_SERVICE_NAME: AZURE_APP_SERVICE_IDENTIFIER_ENV_NAME,
    AZURE_FUNCTION_NAME: AZURE_FUNCTION_IDENTIFIER_ENV_NAME,
    AZURE_ML_SERVICE_NAME: AZURE_ML_ENDPOINT_ENV_NAME,
    AWS_LAMBDA_SERVICE_NAME: AWS_LAMBDA_FUNCTION_IDENTIFIER_ENV_NAME,
    GITHUB_CODESPACE_SERVICE_NAME: GITHUB_CODESPACE_IDENTIFIER_ENV_NAME
}


llm_type_map = {
    "sagemakerendpoint": "aws_sagemaker",
    "azureopenai": "azure_openai",
    "openai": "openai",
    "chatopenai": "openai",
    "azurechatopenai": "azure_openai",
    "bedrock": "aws_bedrock",
    "sagemakerllm": "aws_sagemaker",
    "chatbedrock": "aws_bedrock",
    "openaigenerator": "openai",
    "bedrockruntime":"aws_bedrock",
    "sagemakerruntime":"aws_sagemaker",
    "anthropic": "anthropic",
    "chatanthropic":"anthropic",
    "anthropicchatgenerator":"anthropic",
    "chatcompletionsclient": "azure_ai_inference",
    "embeddingsclient": "azure_ai_inference",
    "imageembeddingsclient": "azure_ai_inference",
    "chatbedrockconverse": "aws_bedrock",
    "googleaigeminichatgenerator": "gemini",
    "gemini": "gemini",
    "chatgooglegenerativeai": "gemini",
}

MONOCLE_INSTRUMENTOR = "monocle_apptrace"
WORKFLOW_TYPE_KEY = "workflow_type"
DATA_INPUT_KEY = "data.input"
DATA_OUTPUT_KEY = "data.output"
PROMPT_INPUT_KEY = "data.input"
PROMPT_OUTPUT_KEY = "data.output"
QUERY = "input"
RESPONSE = "response"
SESSION_PROPERTIES_KEY = "session"
INFRA_SERVICE_KEY = "infra_service_name"
META_DATA = 'metadata'
MONOCLE_SCOPE_NAME_PREFIX = "monocle.scope."
SCOPE_METHOD_LIST = 'MONOCLE_SCOPE_METHODS'
SCOPE_METHOD_FILE = 'monocle_scopes.json'
SCOPE_CONFIG_PATH = 'MONOCLE_SCOPE_CONFIG_PATH'
TRACE_PROPOGATION_URLS = "MONOCLE_TRACE_PROPAGATATION_URLS"
WORKFLOW_TYPE_KEY = "monocle.workflow_type"
ADD_NEW_WORKFLOW = "monocle.add_new_workflow"
WORKFLOW_TYPE_GENERIC = "workflow.generic"
MONOCLE_SDK_VERSION = "monocle_apptrace.version"
MONOCLE_SDK_LANGUAGE = "monocle_apptrace.language"
MONOCLE_DETECTED_SPAN_ERROR = "monocle_apptrace.detected_span_error"
HTTP_SUCCESS_CODES = ('200', '201', '202', '204', '205', '206')
CHILD_ERROR_CODE = "child.error.code"


from monocle_apptrace.instrumentation.metamodel.teamsai import (
    _helper,
)
ACTIONPLANNER_OUTPUT_PROCESSOR = {
    "type": "generic",
    "attributes": [
        [
            {
                "_comment": "planner type and configuration",
                "attribute": "type",
                "accessor": lambda arguments: "teams.planner"
            },
            {
                "attribute": "planner_type",
                "accessor": lambda arguments: "ActionPlanner"
            },
            {
                "attribute": "max_repair_attempts",
                "accessor": lambda arguments: arguments["instance"]._options.max_repair_attempts if hasattr(arguments["instance"], "_options") else 3
            }
        ],
        [
            {
                "_comment": "model configuration",
                "attribute": "model",
                "accessor": lambda arguments: arguments["instance"]._options.model.__class__.__name__ if hasattr(arguments["instance"], "_options") else "unknown"
            },
            {
                "attribute": "tokenizer",
                "accessor": lambda arguments: arguments["instance"]._options.tokenizer.__class__.__name__ if hasattr(arguments["instance"], "_options") else "GPTTokenizer"
            }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "_comment": "input configuration to ActionPlanner",
            "attributes": [
               {
                    "_comment": "execution metadata",
                    "accessor": lambda arguments: _helper.get_prompt_template(arguments)
                },
                {
                    "attribute": "validator",
                    "accessor": lambda arguments: arguments["kwargs"].get("validator").__class__.__name__ if arguments.get("kwargs", {}).get("validator") else "DefaultResponseValidator"
                },
                {
                    "attribute": "memory_type",
                    "accessor": lambda arguments: arguments["kwargs"].get("memory").__class__.__name__ if arguments.get("kwargs", {}).get("memory") else "unknown"
                }
            ]
        },
        {
            "name": "data.output",
            "_comment": "output from ActionPlanner",
            "attributes": [
                {
                    "attribute": "status_check",
                    "accessor": lambda arguments: _helper.status_check(arguments)
                },
                {
                    "attribute": "response",
                    "accessor": lambda arguments: arguments["result"].message.content if hasattr(arguments["result"], "message") else str(arguments["result"])
                }
            ]
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "execution metadata",
                    "accessor": lambda arguments: {
                        "latency_ms": arguments.get("latency_ms"),
                        "feedback_enabled": arguments["instance"]._enable_feedback_loop if hasattr(arguments["instance"], "_enable_feedback_loop") else False
                    }
                }
            ]
        }
    ]
}


"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import json
import logging
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    get_keys_as_tuple,
    get_nested_value,
    get_status_code,
    try_option,
    get_exception_message,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_anthropic_finish_reason_to_finish_type


logger = logging.getLogger(__name__)

def extract_provider_name(instance):
    provider_url: Option[str] = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)

def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if inference_endpoint.is_none() and "meta" in instance.client.__dict__:
        inference_endpoint = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))

def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if "system" in kwargs and isinstance(kwargs["system"], str):
            messages.append({"system": kwargs["system"]})
        if 'messages' in kwargs and len(kwargs['messages']) >0:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})
        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def get_exception_status_code(arguments):
    if arguments['exception'] is not None and hasattr(arguments['exception'], 'status_code') and arguments['exception'].status_code is not None:
        return arguments['exception'].status_code
    elif arguments['exception'] is not None:
        return 'error'
    else:
        return 'success'

def extract_assistant_message(arguments):
    try:
        status = get_status_code(arguments)
        response = arguments["result"]
        if status == 'success':
            messages = []
            role = response.role if hasattr(response, 'role') else "assistant"
            if response is not None and hasattr(response,"content") and len(response.content) >0:
                if hasattr(response.content[0],"text"):
                    messages.append({role: response.content[0].text})
            # return first message if list is not empty
            return get_json_dumps(messages[0]) if messages else ""
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error

    except (IndexError, AttributeError) as e:
        logger.warning("Warning: Error occurred in extract_assistant_message: %s", str(e))
        return None

def update_span_from_llm_response(response):
    meta_dict = {}
    if response is not None and hasattr(response, "usage"):
        if hasattr(response, "usage") and response.usage is not None:
            token_usage = response.usage
        else:
            response_metadata = response.response_metadata
            token_usage = response_metadata.get("token_usage")
        if token_usage is not None:
            meta_dict.update({"completion_tokens": getattr(response.usage, "output_tokens", 0)})
            meta_dict.update({"prompt_tokens": getattr(response.usage, "input_tokens", 0)})
            meta_dict.update({"total_tokens": getattr(response.usage, "input_tokens", 0)+getattr(response.usage, "output_tokens", 0)})
    return meta_dict

def extract_finish_reason(arguments):
    """Extract stop_reason from Anthropic response (Claude)."""
    try:
        # Arguments may be a dict with 'result' or just the response object
        response = arguments.get("result") if isinstance(arguments, dict) else arguments
        if response is not None and hasattr(response, "stop_reason"):
            return response.stop_reason
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    return None

def map_finish_reason_to_finish_type(finish_reason):
    """Map Anthropic stop_reason to finish_type, similar to OpenAI mapping."""
    return map_anthropic_finish_reason_to_finish_type(finish_reason)


from monocle_apptrace.instrumentation.common.wrapper import task_wrapper
from monocle_apptrace.instrumentation.metamodel.flask.entities.http import FLASK_HTTP_PROCESSOR, FLASK_RESPONSE_PROCESSOR

FLASK_METHODS = [
    {
        "package": "flask.app",
        "object": "Flask",
        "method": "wsgi_app",
        "wrapper_method": task_wrapper,
        "span_handler": "flask_handler",
        "output_processor": FLASK_HTTP_PROCESSOR,
    },
    {
        "package": "werkzeug.wrappers.response",
        "object": "Response",
        "method": "__call__",
        "wrapper_method": task_wrapper,
        "span_handler": "flask_response_handler",
        "output_processor": FLASK_RESPONSE_PROCESSOR,
        "skip_span": True
    }
]


from opentelemetry.context import get_value, set_value, attach, detach
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler

class BotoCoreSpanHandler(SpanHandler):

    def _botocore_processor(self, to_wrap, wrapped, instance, args, kwargs, return_value):
        service_name = kwargs.get("service_name")
        service_method_mapping = {
            "sagemaker-runtime": "invoke_endpoint",
            "bedrock-runtime": "converse",
        }
        if service_name in service_method_mapping:
            method_name = service_method_mapping[service_name]
            original_method = getattr(return_value, method_name, None)
            span_name = "botocore-" + service_name + "-invoke-endpoint"
            # wrap_util(original_method, span_name)
            if original_method:
                instrumentor = self.instrumentor
                if instrumentor:
                    instrumented_method = instrumentor(to_wrap, wrapped, span_name, return_value, original_method)
                    setattr(return_value, method_name, instrumented_method)

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value,token=None):
        self._botocore_processor(to_wrap=to_wrap, wrapped=wrapped, instance=instance, return_value=return_value, args=args,
                                 kwargs=kwargs)
        return super().post_tracing(to_wrap, wrapped, instance, args, kwargs,return_value)



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper
from monocle_apptrace.instrumentation.metamodel.mcp import _helper
from monocle_apptrace.instrumentation.metamodel.mcp.entities.inference import TOOLS


MCP_METHODS = [
    {
        "package": "mcp.shared.session",
        "object": "BaseSession",
        "method": "send_request",
        "wrapper_method": atask_wrapper,
        "span_handler": "mcp_agent_handler",
        "output_processor": TOOLS,
    },
    {
        "package": "langchain_mcp_adapters.tools",
        "object": "",
        "method": "convert_mcp_tool_to_langchain_tool",
        "wrapper_method": _helper.langchain_mcp_wrapper,
    },
]



import time, os
import random
import logging
from abc import ABC, abstractmethod
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExportResult
from monocle_apptrace.instrumentation.common.constants import MONOCLE_SDK_VERSION
from typing import Sequence

logger = logging.getLogger(__name__)

class SpanExporterBase(ABC):
    def __init__(self, export_monocle_only: bool = True):
        self.backoff_factor = 2
        self.max_retries = 10
        self.export_queue = []
        self.last_export_time = time.time()
        self.export_monocle_only = export_monocle_only or os.environ.get("MONOCLE_EXPORTS_ONLY", True)

    @abstractmethod
    async def export(self, spans: Sequence[ReadableSpan]) -> SpanExportResult:
        pass

    @abstractmethod
    async def force_flush(self, timeout_millis: int = 30000) -> bool:
        pass

    def shutdown(self) -> None:
        pass

    def skip_export(self, span:ReadableSpan) -> bool:
        if self.export_monocle_only and (not span.attributes.get(MONOCLE_SDK_VERSION)):
            return True
        return False

    @staticmethod
    def retry_with_backoff(retries=3, backoff_in_seconds=1, max_backoff_in_seconds=32, exceptions=(Exception,)):
        def decorator(func):
            def wrapper(*args, **kwargs):
                attempt = 0
                while attempt < retries:
                    try:
                        return func(*args, **kwargs)
                    except exceptions as e:
                        attempt += 1
                        sleep_time = min(max_backoff_in_seconds, backoff_in_seconds * (2 ** (attempt - 1)))
                        sleep_time = sleep_time * (1 + random.uniform(-0.1, 0.1))  # Add jitter
                        logger.warning(f"Network connectivity error, Attempt {attempt} failed: {e}. Retrying in {sleep_time:.2f} seconds...")
                        time.sleep(sleep_time)
                raise Exception(f"Failed after {retries} attempts")

            return wrapper

        return decorator


import logging
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    MonocleSpanException,
    get_keys_as_tuple,
    get_nested_value,
    try_option,
    get_exception_message,
    get_json_dumps,
    get_status_code
)
from monocle_apptrace.instrumentation.metamodel.finish_types import (
    map_teamsai_finish_reason_to_finish_type,
    TEAMSAI_FINISH_REASON_MAPPING
)
from monocle_apptrace.instrumentation.common.constants import CHILD_ERROR_CODE
logger = logging.getLogger(__name__)

def extract_messages(arguments):
    """
    Captures the input from Teams AI state.
    Args:
        arguments (dict): Arguments containing state and context information
    Returns:
        str: The input message or error message
    """
    try:
        # Get the memory object from kwargs
        kwargs = arguments.get("kwargs", {})
        messages = []

        # If memory exists, try to get the input from temp
        if "memory" in kwargs:
            memory = kwargs["memory"]
            # Check if it's a TurnState object
            if hasattr(memory, "get"):
                # Use proper TurnState.get() method
                temp = memory.get("temp")
                if temp and hasattr(temp, "get"):
                    input_value = temp.get("input")
                    if input_value:
                        messages.append({'user': str(input_value)})
        system_prompt = ""
        try:
            system_prompt = kwargs.get("template").prompt.sections[0].sections[0].template
            messages.append({'system': system_prompt})
        except Exception as e:
            print(f"Debug - Error accessing system prompt: {str(e)}")
            
        # Try alternative path through context if memory path fails
        context = kwargs.get("context")
        if hasattr(context, "activity") and hasattr(context.activity, "text"):
            messages.append({'user': str(context.activity.text)})

        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        print(f"Debug - Arguments structure: {str(arguments)}")
        print(f"Debug - kwargs: {str(kwargs)}")
        if "memory" in kwargs:
            print(f"Debug - memory type: {type(kwargs['memory'])}")
        return f"Error capturing input: {str(e)}"

def capture_prompt_info(arguments):
    """Captures prompt information from ActionPlanner state"""
    try:
        kwargs = arguments.get("kwargs", {})
        prompt = kwargs.get("prompt")

        if isinstance(prompt, str):
            return prompt
        elif hasattr(prompt, "name"):
            return prompt.name

        return "No prompt information found"
    except Exception as e:
        return f"Error capturing prompt: {str(e)}"

def capture_prompt_template_info(arguments):
    """Captures prompt information from ActionPlanner state"""
    try:
        kwargs = arguments.get("kwargs", {})
        prompt = kwargs.get("prompt")

        if hasattr(prompt,"prompt") and prompt.prompt is not None:
            if "_text" in prompt.prompt.__dict__:
                prompt_template = prompt.prompt.__dict__.get("_text", None)
                return prompt_template
            elif "_sections" in prompt.prompt.__dict__ and prompt.prompt._sections is not None:
                sections = prompt.prompt._sections[0].__dict__.get("_sections", None)
                if sections is not None and "_template" in sections[0].__dict__:
                    return sections[0].__dict__.get("_template", None)


        return "No prompt information found"
    except Exception as e:
        return f"Error capturing prompt: {str(e)}"

def status_check(arguments):
    if hasattr(arguments["result"], "error") and arguments["result"].error is not None:
        error_msg:str = arguments["result"].error
        error_code:str = arguments["result"].status if hasattr(arguments["result"], "status") else "unknown"
        raise MonocleSpanException(f"Error: {error_code} - {error_msg}")

def get_prompt_template(arguments):
    pass
    return {
        "prompt_template_name": capture_prompt_info(arguments),
        "prompt_template": capture_prompt_template_info(arguments),
        "prompt_template_description": get_nested_value(arguments.get("kwargs", {}), ["prompt", "config", "description"]),
        "prompt_template_type": get_nested_value(arguments.get("kwargs", {}), ["prompt", "config", "type"])
    }

def get_status(arguments):
    if arguments["exception"] is not None:
        return 'error'
    elif get_status_code(arguments) == 'success':
        return 'success'
    else:
        return 'error'

def extract_assistant_message(arguments) -> str:
    status = get_status_code(arguments)
    messages = []
    role = "assistant"
    if status == 'success':
        if hasattr(arguments["result"], "message"):
            messages.append({role: arguments["result"].message.content})
        else:
            messages.append({role: str(arguments["result"])})
    else:
        if arguments["exception"] is not None:
            return get_exception_message(arguments)
        elif hasattr(arguments["result"], "error"):
            return arguments["result"].error
    return get_json_dumps(messages[0]) if messages else ""

def extract_finish_reason(arguments):
    """Map TeamAI finish_reason to standardized finish_type."""
    return get_status_code(arguments)

def extract_status_code(arguments):
    # TeamsAI doesn't capture the status and other metadata from underlying OpenAI SDK.
    # Thus we save the OpenAI status code in the parent span and retrieve it here to preserve meaningful error codes.
    status = get_status_code(arguments)
    if status == 'error' and arguments['exception'] is None:
        child_status = arguments['span'].attributes.get(CHILD_ERROR_CODE)
        if child_status is not None:
            return child_status
    return status

def check_status(arguments):
    status = get_status_code(arguments)
    if status != 'success' and arguments['exception'] is None:
        raise MonocleSpanException(f"{status}")   

def map_finish_reason_to_finish_type(finish_reason):
    """Map TeamsAI finish_reason to standardized finish_type."""
    if not finish_reason:
        return None

    return map_teamsai_finish_reason_to_finish_type(finish_reason)

def extract_provider_name(instance):
    provider_url: Option[str] = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)


def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if inference_endpoint.is_none() and "meta" in instance.client.__dict__:
        inference_endpoint = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))


from monocle_apptrace.instrumentation.common.wrapper import task_wrapper
from monocle_apptrace.instrumentation.metamodel.haystack.entities.inference import INFERENCE
from monocle_apptrace.instrumentation.metamodel.haystack.entities.retrieval import RETRIEVAL

HAYSTACK_METHODS = [
    {
        "package": "haystack.components.retrievers.in_memory",
        "object": "InMemoryEmbeddingRetriever",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": RETRIEVAL
    },
    {
        "package": "haystack_integrations.components.retrievers.opensearch",
        "object": "OpenSearchEmbeddingRetriever",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": RETRIEVAL
    },
    {
        "package": "haystack.components.generators.openai",
        "object": "OpenAIGenerator",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "haystack.components.generators.chat.openai",
        "object": "OpenAIChatGenerator",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "haystack.core.pipeline.pipeline",
        "object": "Pipeline",
        "method": "run",
        "wrapper_method": task_wrapper
    },
    {
        "package": "haystack_integrations.components.generators.anthropic",
        "object": "AnthropicChatGenerator",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "haystack_integrations.components.generators.google_ai",
        "object": "GoogleAIGeminiChatGenerator",
        "method": "run",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
]



from abc import ABC, abstractmethod
import logging
import os
import queue
import threading
import time
from typing import Callable
import requests
from monocle_apptrace.instrumentation.common.constants import AWS_LAMBDA_ENV_NAME

logger = logging.getLogger(__name__)
LAMBDA_EXTENSION_NAME = "AsyncProcessorMonocle"

class ExportTaskProcessor(ABC):
    
    @abstractmethod
    def start(self):
        return

    @abstractmethod
    def stop(self):
        return

    @abstractmethod
    def queue_task(self, async_task: Callable[[Callable, any], any] = None, args: any = None, is_root_span: bool = False):
        return

class LambdaExportTaskProcessor(ExportTaskProcessor):
    
    def __init__(
        self,
        span_check_interval_seconds: int = 1,
        max_time_allowed_seconds: int = 30):
        # An internal queue used by the handler to notify the extension that it can
        # start processing the async task.
        self.async_tasks_queue = queue.Queue()
        self.span_check_interval = span_check_interval_seconds
        self.max_time_allowed = max_time_allowed_seconds

    def start(self):
        try:
            self._start_async_processor()
        except Exception as e:
            logger.error(f"LambdaExportTaskProcessor| Failed to start. {e}")

    def stop(self):
        return

    def queue_task(self, async_task=None, args=None, is_root_span=False):
        self.async_tasks_queue.put((async_task, args, is_root_span))
    
    def set_sagemaker_model(self, endpoint_name: str, span: dict[str, dict[str, str]]):
        try:
            try:
                import boto3
            except ImportError:
                logger.error("LambdaExportTaskProcessor| Failed to import boto3")
                return
            
            client = boto3.client('sagemaker')
            response = client.describe_endpoint(
                EndpointName=endpoint_name
            )
            endpoint_config_name = response["EndpointConfigName"]
            endpoint_config_response = client.describe_endpoint_config(
                EndpointConfigName=endpoint_config_name
            )
            model_name = endpoint_config_response["ProductionVariants"][0]["ModelName"]
            model_name_response = client.describe_model(ModelName = model_name)
            model_name_id = ""
            try:
                model_name_id = model_name_response["PrimaryContainer"]["Environment"]["HF_MODEL_ID"]
            except:
                pass
            span["attributes"]["model_name"] = model_name_id
        except Exception as e:
            logger.error(f"LambdaExportTaskProcessor| Failed to get sagemaker model. {e}")

    def update_spans(self, export_args):
        try:
            if 'batch' in export_args:
                for span in export_args["batch"]:
                    try:
                        if len(span["attributes"]["sagemaker_endpoint_name"]) > 0 :
                            self.set_sagemaker_model(endpoint_name=span["attributes"]["sagemaker_endpoint_name"], span=span)
                    except:
                        pass
        except Exception as e:
            logger.error(f"LambdaExportTaskProcessor| Failed to update spans. {e}")

    def _start_async_processor(self):
        # Register internal extension
        logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Registering with Lambda service...")
        response = requests.post(
            url=f"http://{os.environ['AWS_LAMBDA_RUNTIME_API']}/2020-01-01/extension/register",
            json={'events': ['INVOKE']},
            headers={'Lambda-Extension-Name': LAMBDA_EXTENSION_NAME}
        )
        ext_id = response.headers['Lambda-Extension-Identifier']
        logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Registered with ID: {ext_id}")

        def process_tasks():
            while True:
                # Call /next to get notified when there is a new invocation and let
                # Lambda know that we are done processing the previous task.

                logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Waiting for invocation...")
                response = requests.get(
                    url=f"http://{os.environ['AWS_LAMBDA_RUNTIME_API']}/2020-01-01/extension/event/next",
                    headers={'Lambda-Extension-Identifier': ext_id},
                    timeout=None
                )
                root_span_found = False
                # all values in seconds
                total_time_elapsed = 0
                while root_span_found is False and total_time_elapsed < self.max_time_allowed:
                    logger.debug(response.json())
                    # Get next task from internal queue
                    logger.info(f"[{LAMBDA_EXTENSION_NAME}] Async thread running, waiting for task from handler")
                    while self.async_tasks_queue.empty() is False :
                        logger.info(f"[{LAMBDA_EXTENSION_NAME}] Processing task from handler")
                        async_task, arg, is_root_span = self.async_tasks_queue.get()
                        root_span_found = is_root_span
                        # self.update_spans(export_args=arg)

                        if async_task is None:
                            # No task to run this invocation
                            logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Received null task. Ignoring.")
                        else:
                            # Invoke task
                            logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Received async task from handler. Starting task.")
                            async_task(arg)
                    total_time_elapsed+=self.span_check_interval
                    logger.info(f"[{LAMBDA_EXTENSION_NAME}] Waiting for root span. total_time_elapsed: {total_time_elapsed}, root_span_found: {root_span_found}.")
                    time.sleep(self.span_check_interval)
                
                logger.debug(f"[{LAMBDA_EXTENSION_NAME}] Finished processing task. total_time_elapsed: {total_time_elapsed}, root_span_found: {root_span_found}.")

        # Start processing extension events in a separate thread
        threading.Thread(target=process_tasks, daemon=True, name=LAMBDA_EXTENSION_NAME).start() 


def is_aws_lambda_environment():
    return AWS_LAMBDA_ENV_NAME in os.environ


from monocle_apptrace.instrumentation.metamodel.langgraph import (
    _helper
)

AGENT = {
      "type": "agentic.invocation",
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.langgraph'
              },
              {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_agent_name(arguments['instance'])
              },
              {
                "_comment": "agent description",
                "attribute": "description",
                "accessor": lambda arguments: _helper.get_agent_description(arguments['instance'])
              }
        ]
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Agent input",
                "attribute": "query",
                "accessor": lambda arguments: _helper.extract_agent_input(arguments)
            }
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from LLM",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_agent_response(arguments['result'])
            }
          ]
        }
      ]
    }

AGENT_REQUEST = {
      "type": "agentic.request",
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.langgraph'
              }
        ],
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Agent input",
                "attribute": "input",
                "accessor": lambda arguments: _helper.extract_agent_input(arguments)
            }
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from LLM",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_agent_response(arguments['result'])
            }
          ]
        }
      ]
}

TOOLS = {
      "type": "agentic.tool.invocation",
      "attributes": [
        [
              {
                "_comment": "tool type",
                "attribute": "type",
                "accessor": lambda arguments:'tool.langgraph'
              },
              {
                "_comment": "name of the tool",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_tool_name(arguments['instance'])
              },
              {
                  "_comment": "tool description",
                  "attribute": "description",
                  "accessor": lambda arguments: _helper.get_tool_description(arguments['instance'])
              }
        ],
        [             
              {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_source_agent()
              },          
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.langgraph'
              }
        ]
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Tool input",
                "attribute": "Inputs",
                "accessor": lambda arguments: _helper.extract_tool_input(arguments) 
            },
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from Tool",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_tool_response(arguments['result'])
            }
          ]
        }
      ]
}

AGENT_DELEGATION = {
      "type": "agentic.delegation",
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.langgraph'
              },
              {
                "_comment": "name of the agent",
                "attribute": "from_agent",
                "accessor": lambda arguments: _helper.get_source_agent()
              },
              {
                "_comment": "name of the agent called",
                "attribute": "to_agent",
                "accessor": lambda arguments: _helper.get_target_agent(arguments['instance'])
              }
        ]
      ]
}


from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import resolve_from_alias
import logging
logger = logging.getLogger(__name__)

DELEGATION_NAME_PREFIX = 'transfer_to_'
ROOT_AGENT_NAME = 'LangGraph'
LANGGRAPTH_AGENT_NAME_KEY = "agent.langgraph"

def extract_agent_response(response):
    try:
        if response is not None and 'messages' in response:
            output = response["messages"][-1]
            return str(output.content)
    except Exception as e:
        logger.warning("Warning: Error occurred in handle_response: %s", str(e))
    return ""

def agent_instructions(arguments):
    if callable(arguments['kwargs']['agent'].instructions):
        return arguments['kwargs']['agent'].instructions(arguments['kwargs']['context_variables'])
    else:
        return arguments['kwargs']['agent'].instructions

def extract_agent_input(arguments):
    if arguments['result'] is not None and 'messages' in arguments['result']:
        history = arguments['result']['messages']
        for message in history:
            if hasattr(message, 'content') and hasattr(message, 'type') and message.type == "human":  # Check if the message is a HumanMessage
                return message.content
    return None

def get_inference_endpoint(arguments):
    inference_endpoint = resolve_from_alias(arguments['instance'].client.__dict__, ['azure_endpoint', 'api_base', '_base_url'])
    return str(inference_endpoint)

def tools(instance):
    if hasattr(instance,'nodes') and ('tools' in instance.nodes):
        tools= instance.nodes['tools']
        if hasattr(tools,'bound') and hasattr(tools.bound,'tools_by_name'):
            return list(tools.bound.tools_by_name.keys())

def update_span_from_llm_response(response):
    meta_dict = {}
    token_usage = None
    if response is not None and "messages" in response:
        token = response["messages"][-1]
        if token.response_metadata is not None:
            token_usage = token.response_metadata["token_usage"]
        if token_usage is not None:
            meta_dict.update({"completion_tokens": token_usage.get('completion_tokens')})
            meta_dict.update({"prompt_tokens": token_usage.get('prompt_tokens')})
            meta_dict.update({"total_tokens": token_usage.get('total_tokens')})
    return meta_dict

def extract_tool_response(result):
    if result is not None and hasattr(result, 'content'):
        return result.content
    return None

def get_status(result):
    if result is not None and hasattr(result, 'status'):
        return result.status
    return None

def extract_tool_input(arguments):
    tool_input = arguments['args'][0]
    if isinstance(tool_input, str):
        return [tool_input]
    else:
        return list(tool_input.values())

def get_name(instance):
    return instance.name if hasattr(instance, 'name') else ""

def get_agent_name(instance) -> str:
    return get_name(instance)

def get_tool_name(instance) -> str:
    return get_name(instance)

def is_delegation_tool(instance) -> bool:
    return get_name(instance).startswith(DELEGATION_NAME_PREFIX)

def get_target_agent(instance) -> str:
    return get_name(instance).replace(DELEGATION_NAME_PREFIX, '', 1)

def is_root_agent_name(instance) -> bool:
    return get_name(instance) == ROOT_AGENT_NAME

def get_source_agent() -> str:
    """Get the name of the agent that initiated the request."""
    from_agent = get_value(LANGGRAPTH_AGENT_NAME_KEY)
    return from_agent if from_agent is not None else ""

def get_description(instance) -> str:
    return instance.description if hasattr(instance, 'description') else ""

def get_agent_description(instance) -> str:
    """Get the description of the agent."""
    return get_description(instance)

def get_tool_description(instance) -> str:
    """Get the description of the tool."""
    return get_description(instance)


from opentelemetry.context import attach, detach, get_current, get_value, set_value, Context
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.metamodel.llamaindex._helper import (
    is_delegation_tool, LLAMAINDEX_AGENT_NAME_KEY, get_agent_name
)
from monocle_apptrace.instrumentation.metamodel.llamaindex.entities.agent import (
    AGENT_DELEGATION
)

TOOL_INVOCATION_STARTED:str = "llamaindex.tool_invocation_started"

class DelegationHandler(SpanHandler):
    # LlamaIndex uses an internal tool to initate delegation to other agents. The method is tool invoke() with tool name as `transfer_to_<agent_name>`.
    # Hence we usea different output processor for tool invoke() to format the span as agentic.delegation.
    def hydrate_span(self, to_wrap, wrapped, instance, args, kwargs, result, span, parent_span = None, ex:Exception = None) -> bool:
        if is_delegation_tool(args, instance):
            agent_request_wrapper = to_wrap.copy()
            agent_request_wrapper["output_processor"] = AGENT_DELEGATION
        else:
            agent_request_wrapper = to_wrap

        return super().hydrate_span(agent_request_wrapper, wrapped, instance, args, kwargs, result, span, parent_span, ex)


# There are two different APIs for tool calling FunctionTool.call() and AgentWorkflow.tool_call(). In case of single agent calling tool, only the FunctionTool.call() is used. In case of multi agent case,
# the AgentWorkflow.tool_call() is used which inturn calls FunctionTool.call(). We can't entirely rely on the FunctionTool.call() to extract tool span details, especially the agent delegation details are not available there.
# Hence we want to distinguish between single agent tool call and multi agent tool call. In case of multi agent tool call, we suppress the FunctionTool.call() span and use AgentWorkflow.tool_call() span to capture the tool call details.
class LlamaIndexToolHandler(DelegationHandler):
    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return attach(set_value(TOOL_INVOCATION_STARTED, True))
    
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value, token=None):
        if token:
            detach(token)

class LlamaIndexSingleAgenttToolHandlerWrapper(DelegationHandler):
    def skip_span(self, to_wrap, wrapped, instance, args, kwargs):
        if get_value(TOOL_INVOCATION_STARTED) == True:
            return True
        return super().skip_span(to_wrap, wrapped, instance, args, kwargs)

class LlamaIndexAgentHandler(SpanHandler):
    # LlamaIndex uses direct OpenAI call for agent inferences. Given that the workflow type is set to llamaindex, the openAI inference does not record the input/output events.
    # To avoid this, we set the workflow type to generic for agent inference spans so we can capture the prompts and responses.
    def hydrate_span(self, to_wrap, wrapped, instance, args, kwargs, result, span, parent_span = None, ex:Exception = None) -> bool:
        retval = super().hydrate_span(to_wrap, wrapped, instance, args, kwargs, result, span, parent_span, ex)
        if SpanHandler.is_root_span(parent_span):
            span.set_attribute(LLAMAINDEX_AGENT_NAME_KEY, "")
        else:
            parent_span.set_attribute(LLAMAINDEX_AGENT_NAME_KEY, get_agent_name(instance))
        return retval



from monocle_apptrace.instrumentation.metamodel.fastapi import _helper

FASTAPI_HTTP_PROCESSOR = {
    "type": "http.process",
    "attributes": [
        [
            {
                "attribute": "method",
                "accessor": lambda arguments: _helper.get_method(arguments['args'][0])
            },
            {
                "attribute": "route",
                "accessor": lambda arguments: _helper.get_route(arguments['args'][0])
            },
        ]
    ]
}

FASTAPI_RESPONSE_PROCESSOR = {
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "attribute": "params",
                    "accessor": lambda arguments: _helper.get_params(arguments['args'][0])
                }
            ]
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "status",
                    "accessor": lambda arguments: _helper.extract_status(arguments['instance'])
                },
                {
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_response(arguments['instance'])
                }
            ]
        }
    ]
}


from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.azureaiinference.entities.inference import INFERENCE

AZURE_AI_INFERENCE_METHODS = [
    # Chat Completions - Synchronous
    {
        "package": "azure.ai.inference",
        "object": "ChatCompletionsClient",
        "method": "complete",
        "wrapper_method": task_wrapper,
        "span_handler": "non_framework_handler",
        "output_processor": INFERENCE
    },
    # Chat Completions - Asynchronous
    {
        "package": "azure.ai.inference.aio",
        "object": "ChatCompletionsClient",
        "method": "complete",
        "wrapper_method": atask_wrapper,
        "span_handler": "non_framework_handler",
        "output_processor": INFERENCE
    }
]



from monocle_apptrace.instrumentation.metamodel.teamsai import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import get_error_message, get_llm_type
TEAMAI_OUTPUT_PROCESSOR = {
    "type": "inference.framework",
    "attributes": [
        [
            {
                "_comment": "provider type, name, deployment",
                "attribute": "type",
                "accessor": lambda arguments: 'inference.' + (get_llm_type(arguments['instance']._client) or 'generic')
            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: _helper.extract_provider_name(arguments['instance'])
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: _helper.extract_inference_endpoint(arguments['instance'])
            }
        ],
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: arguments["instance"]._options.default_model if hasattr(arguments["instance"], "_options") else "unknown"
            },
            {
                "_comment": "LLM Model",
                "attribute": "type",
                "accessor": lambda arguments: 'model.llm.'+ arguments["instance"]._options.default_model if hasattr(arguments["instance"], "_options") else "unknown"
            },
            {
                "attribute": "is_streaming",
                "accessor": lambda arguments: arguments["instance"]._options.stream if hasattr(arguments["instance"], "_options") else False
            }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "_comment": "input to Teams AI",
            "attributes": [
                {
                    "attribute": "input",
                    "accessor": _helper.extract_messages
                }
            ]
        },
        {
            "name": "data.output",
            "_comment": "output from Teams AI",
            "attributes": [
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: _helper.extract_status_code(arguments)
                },
                {
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments)
                },
                {
                    "attribute": "check_status",
                    "accessor": lambda arguments: _helper.check_status(arguments)
                }
            ]
        },
    ]
}


from monocle_apptrace.instrumentation.metamodel.a2a import _helper

A2A_CLIENT = {
    "type": "agentic.invocation",
    "attributes": [
        [
            {
                "attribute": "type",
                "accessor": lambda arguments: "agent2agent.server"
            },
            {
                "attribute": "url",
                "accessor": lambda arguments: _helper.get_url(arguments)
            },
            {
                "attribute": "method",
                "accessor": lambda arguments: _helper.get_method(arguments)
            }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is a2a input",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.get_params_arguments(arguments)
                },
                {
                    "_comment": "this is a2a input",
                    "attribute": "role",
                    "accessor": lambda arguments: _helper.get_role(arguments)
                },
            ],
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is a2a output",
                    "attribute": "status",
                    "accessor": lambda arguments: _helper.get_status(arguments, "status")
                },
                {
                    "_comment": "this is a2a output",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.get_response(arguments)   
                },
            ],
        },
    ],
}

# A2A_RESOLVE = {
#     "type": "a2a.resolve",
#     "attributes": [
#         [
#             # {
#             #   "_comment": "tool type",
#             #   "attribute": "type",
#             #   "accessor": lambda arguments:'tool.mcp'
#             # },
#             {
#                 "_comment": "name of the tool",
#                 "attribute": "name",
#                 "accessor": lambda arguments: _helper.log(arguments),
#             },
#                         {
#                 "_comment": "tool description",
#                 "attribute": "agent_description",
#                 "accessor": lambda arguments: arguments["result"].description
#             },
#             {
#                 "_comment": "tool name",
#                 "attribute": "agent_name",
#                 "accessor": lambda arguments: arguments["result"].name
#             }
#             # {
#             #     "_comment": "tool type",
#             #     "attribute": "type",
#             #     "accessor": lambda arguments: _helper.get_type(arguments),
#             # },
#         ]
#     ],
#     "events": [
#         # {
#         #     "name": "data.input",
#         #     "attributes": [
#         #         {
#         #             "_comment": "this is Tool input",
#         #             "attribute": "input",
#         #             "accessor": lambda arguments: _helper.get_params_arguments(
#         #                 arguments
#         #             ),
#         #         },
#         #     ],
#         # },
#         # {
#         #     "name": "data.output",
#         #     "attributes": [
#         #         {
#         #             "_comment": "this is Tool output",
#         #             "attribute": "output",
#         #             "accessor": lambda arguments: _helper.get_output_text(arguments)
#         #         },
#         #     ],
#         # },
#     ],
# }





import logging
import time
from types import SimpleNamespace
from monocle_apptrace.instrumentation.metamodel.azureaiinference import _helper
from monocle_apptrace.instrumentation.common.utils import (
    get_error_message,
    resolve_from_alias, 
    patch_instance_method,
    get_status,
    get_exception_status_code
)

logger = logging.getLogger(__name__)


def process_stream(to_wrap, response, span_processor):
    """Process streaming responses from Azure AI Inference."""
    waiting_for_first_token = True
    stream_start_time = time.time_ns()
    first_token_time = stream_start_time
    stream_closed_time = None
    accumulated_response = ""
    token_usage = None
    role = "assistant"

    # For sync iteration - patch __next__ instead of __iter__
    if to_wrap and hasattr(response, "__next__"):
        original_next = response.__next__

        def new_next(self):
            nonlocal waiting_for_first_token, first_token_time, stream_closed_time, accumulated_response, token_usage, role

            try:
                item = original_next()
                
                # Handle Azure AI Inference streaming chunks
                if hasattr(item, 'choices') and item.choices:
                    choice = item.choices[0]
                    if hasattr(choice, 'delta') and hasattr(choice.delta, 'role') and choice.delta.role:
                        role = choice.delta.role
                    if hasattr(choice, 'delta') and hasattr(choice.delta, 'content') and choice.delta.content:
                        if waiting_for_first_token:
                            waiting_for_first_token = False
                            first_token_time = time.time_ns()

                        accumulated_response += choice.delta.content
                
                # Check for usage information at the end of stream
                if hasattr(item, 'usage') and item.usage:
                    token_usage = item.usage
                    stream_closed_time = time.time_ns()

                return item

            except StopIteration:
                # Stream is complete, process final span
                if span_processor:
                    ret_val = SimpleNamespace(
                        type="stream",
                        role=role,
                        timestamps={
                            "data.input": int(stream_start_time),
                            "data.output": int(first_token_time),
                            "metadata": int(stream_closed_time or time.time_ns()),
                        },
                        output_text=accumulated_response,
                        usage=token_usage,
                    )
                    span_processor(ret_val)
                raise
            except Exception as e:
                logger.warning(
                    "Warning: Error occurred while processing item in new_next: %s",
                    str(e),
                )
                raise

        patch_instance_method(response, "__next__", new_next)
        
    # For async iteration - patch __anext__ instead of __aiter__
    if to_wrap and hasattr(response, "__anext__"):
        original_anext = response.__anext__

        async def new_anext(self):
            nonlocal waiting_for_first_token, first_token_time, stream_closed_time, accumulated_response, token_usage, role

            try:
                item = await original_anext()
                
                # Handle Azure AI Inference streaming chunks
                if hasattr(item, 'choices') and item.choices:
                    choice = item.choices[0]
                    if hasattr(choice, 'delta') and hasattr(choice.delta, 'role') and choice.delta.role:
                        role = choice.delta.role
                    if hasattr(choice, 'delta') and hasattr(choice.delta, 'content') and choice.delta.content:
                        if waiting_for_first_token:
                            waiting_for_first_token = False
                            first_token_time = time.time_ns()

                        accumulated_response += choice.delta.content
                
                # Check for usage information at the end of stream
                if hasattr(item, 'usage') and item.usage:
                    token_usage = item.usage
                    stream_closed_time = time.time_ns()

                return item

            except StopAsyncIteration:
                # Stream is complete, process final span
                if span_processor:
                    ret_val = SimpleNamespace(
                        type="stream",
                        role=role,
                        timestamps={
                            "data.input": int(stream_start_time),
                            "data.output": int(first_token_time),
                            "metadata": int(stream_closed_time or time.time_ns()),
                        },
                        output_text=accumulated_response,
                        usage=token_usage,
                    )
                    span_processor(ret_val)
                raise
            except Exception as e:
                logger.warning(
                    "Warning: Error occurred while processing item in new_anext: %s",
                    str(e),
                )
                raise

        patch_instance_method(response, "__anext__", new_anext)


INFERENCE = {
    "type": "inference",
    "is_auto_close": lambda kwargs: kwargs.get("stream", False) is False,
    "response_processor": process_stream,
    "attributes": [
        [
            {
                "_comment": "Azure AI Inference provider type, endpoint",
                "attribute": "type",
                "accessor": lambda arguments: f"inference.{_helper.get_inference_type(arguments)}"
            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: _helper.get_provider_name(arguments['instance'])
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: _helper.extract_inference_endpoint(arguments['instance'])
            },
            {
                "attribute": "deployment",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments['instance'].__dict__,
                    ['deployment', 'deployment_name', 'azure_deployment', '_deployment']
                )
            }
        ],
        [
            {
                "_comment": "LLM Model information",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_model_name(arguments)
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: f"model.llm.{_helper.get_model_name(arguments)}" if _helper.get_model_name(arguments) else "model.llm.unknown"
            }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "Chat messages input to Azure AI Inference",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_messages(arguments['kwargs'])
                }
            ]
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "Response from Azure AI Inference",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments)
                },
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                }
            ]
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "Usage metadata from Azure AI Inference",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(
                        arguments['result'], 
                        arguments.get('instance')
                    )
                },
                {
                    "attribute": "finish_reason",
                    "accessor": lambda arguments: _helper.extract_finish_reason(arguments)
                },
                {
                    "attribute": "finish_type",
                    "accessor": lambda arguments: _helper.map_finish_reason_to_finish_type(
                        _helper.extract_finish_reason(arguments)
                    )
                }
            ]
        }
    ]
}



import logging
from threading import local
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes, try_option, Option, \
    MonocleSpanException
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from urllib.parse import unquote, urlparse, ParseResult

logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000
token_data = local()
token_data.current_token = None

def get_url(kwargs) -> ParseResult:
    url_str = try_option(lambda k: k.get('path'), kwargs['event'])
    url = url_str.unwrap_or(None)
    if url is not None:
        return urlparse(url)
    else:
        return None

def get_route(args) -> str:
    event = args[1]
    route = event.get("path") or event.get("requestContext", {}).get("path")
    return route

def get_method(args) -> str:
    event = args[1]
    http_method = event.get("httpMethod") or event.get("requestContext", {}).get("httpMethod")
    return http_method


def get_params(args) -> dict:
    event = args[1]
    question = None
    query_params = event.get('queryStringParameters', {})
    if isinstance(query_params, dict):
        question = query_params.get('question')
    return question

def get_body(args) -> dict:
    event = args[1]
    body = event.get("body")
    return body

def extract_response(result) -> str:
    if isinstance(result, dict) and 'body' in result:
        response = result['body']
        if isinstance(response, bytes):
            response = response.decode('utf-8', errors='ignore')
    else:
        response = ""
    return response


def extract_status(result) -> str:
    status = f"{result['statusCode']}" if isinstance(result, dict) and 'statusCode' in result else ""
    if status not in HTTP_SUCCESS_CODES:
        error_message = extract_response(result)
        raise MonocleSpanException(f"error: {status} - {error_message}")
    return status


def lambda_func_pre_tracing(kwargs):
    headers = kwargs['event'].get('headers', {}) if 'event' in kwargs else {}
    return extract_http_headers(headers)


def lambda_func_post_tracing(token):
    clear_http_scopes(token)


class lambdaSpanHandler(SpanHandler):
    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return lambda_func_pre_tracing(kwargs)

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value,token):
        lambda_func_post_tracing(token)
